from __future__ import with_statement

import os
import time
import subprocess

from decimal import Decimal

from fabric.api import *
from fabric.contrib import files

from fabulaws.api import call_python
from fabulaws.decorators import uses_fabric
from fabulaws.ec2 import EC2Instance
from fabulaws.ubuntu.instances import UbuntuInstance
from fabulaws.ubuntu.packages.fail2ban import Fail2banMixin
from fabulaws.ubuntu.packages.postgres import PostgresMixin
from fabulaws.ubuntu.packages.redis import RedisMixin
from fabulaws.ubuntu.packages.memcached import MemcachedMixin
from fabulaws.ubuntu.packages.python import PythonMixin
from fabulaws.ubuntu.packages.rabbitmq import RabbitMqMixin


__all__ = [
    'CacheInstance',
    'DbMasterInstance',
    'DbSlaveInstance',
    'WorkerInstance',
    'WebInstance',
    'CombinedInstance',
]


class FirewallMixin(Fail2banMixin):
    pass


class BaseInstance(FirewallMixin, UbuntuInstance):
    deployment_dir = os.path.dirname(__file__)
    project_root = os.path.dirname(deployment_dir)

    def __init__(self, *args, **kwargs):
        for key in ['ami', 'key_prefix', 'admin_groups', 'run_upgrade',
                    'app_root', 'deploy_user_home', 'fs_type', 'fs_encrypt',
                    'ubuntu_mirror', 'swap_multiplier', 'instance_type',
                    'deploy_user', 'volume_size', 'volume_type',
                    'security_groups']:
            setattr(self, key, kwargs.pop(key, ''))
        if 'terminate' not in kwargs:
            kwargs['terminate'] = False
        passwd = getattr(env, 'luks_passphrase', None)
        self.volume_info = [('/dev/sdf', self.app_root, self.volume_size,
                             self.volume_type, passwd)]
        self.default_swap_file = '%s/swapfile' % self.app_root
        super(BaseInstance, self).__init__(*args, **kwargs)
        self.app_dirs = ['/tmp', self.deploy_user_home] # mixins add other dirs

    def _get_users(self):
        """
        Returns a list of tuples of (username, key_file_path).
        """
        users_dir = env.ssh_keys
        users = [(n, os.path.join(users_dir, n))
                 for n in os.listdir(users_dir)
                 if '.' not in n]  # skip files ending in '.gecos' or other suffixes
        return users

    def _add_swap(self, path):
        sudo('mkswap -f {0}'.format(path))
        with settings(warn_only=True):
            # sometimes mkswap seems to 'mount' the swap partition
            # automatically, so this command will fail
            sudo('swapon {0}'.format(path))
        files.append('/etc/fstab', '{0} none swap sw 0 0'.format(path), use_sudo=True)

    @uses_fabric
    def setup_swap(self):
        """Sets up swap partition"""
        swap_mb = self.server_memory * self.swap_multiplier
        devs = ['/dev/xvdb', '/dev/xvdc']
        for dev in devs:
            print 'attempting swap creation on {0}...'.format(dev)
            if swap_mb <= 0:
                print 'no additional swap needed; skipping'
                break
            if not files.exists(dev):
                print 'no such device {0}; skipping'.format(dev)
                continue
            size = sudo('blockdev --getsize64 {0}'.format(dev))
            try:
                size = int(size) / 1024 / 1024
            except ValueError:
                print 'no size found for {0}'.format(dev)
                continue
            # decrement size regardless of whether or not we create swap; if
            # crypt device exists, assume we created it previously and don't 
            # create even more swap
            swap_mb -= size
            with settings(warn_only=True):
                sudo('umount {0}'.format(dev))
            crypt_name = 'cryptswap-{0}'.format(dev.split('/')[-1])
            if files.exists('/dev/mapper/{0}'.format(crypt_name)):
                print 'crypt device {0} already exists; skipping'.format(crypt_name)
                continue
            sudo('cryptsetup -d /dev/urandom create {0} {1}'.format(crypt_name, dev))
            files.append('/etc/crypttab',
                         '{0} {1} /dev/urandom swap'.format(crypt_name, dev),
                         use_sudo=True)
            files.comment('/etc/fstab', dev, use_sudo=True)
            self._add_swap('/dev/mapper/{0}'.format(crypt_name))
        if swap_mb > 0:
            if not files.exists(self.default_swap_file):
                sudo('dd if=/dev/zero of={0} bs=1M count={1}'.format(self.default_swap_file, swap_mb))
                sudo('chown root:root {0}'.format(self.default_swap_file))
                sudo('chmod 600 {0}'.format(self.default_swap_file))
                self._add_swap(self.default_swap_file)
            else:
                print 'swap file {0} already exists; skipping'.format(self.default_swap_file)
        # remove old swap at the end in case it's already in use (otherwise swapoff might fail)
        old_swap = ['/dev/xvda3']
        for swap in old_swap:
            if files.contains('/proc/swaps', swap):
                sudo('swapoff {0}'.format(swap))
                files.comment('/etc/fstab', swap, use_sudo=True)
                with settings(warn_only=True):
                    sudo('dd if=/dev/zero of={0} bs=1M'.format(swap))

    @uses_fabric
    def setup_sudoers(self):
        """
        Creates the sudoers file on the server, based on the supplied template.
        """
        sudoers_file = os.path.join(self.deployment_dir, 'templates', 'sudoers')
        files.upload_template(sudoers_file, '/etc/sudoers.new', backup=False,
                              use_sudo=True, mode=0440)
        sudo('chown root:root /etc/sudoers.new')
        sudo('mv /etc/sudoers.new /etc/sudoers')

    def reset_authentication(self):
        """
        Delete's the 'ubuntu' user in the AMI once it's no longer needed.
        """
        user = self.user  # save user before it's overwritten by super() method
        super(BaseInstance, self).reset_authentication()
        if user == 'ubuntu':
            with self:
                # disable the 'ubuntu' user so it can no longer log in
                sudo('usermod --expiredate 1 {}'.format(user))

    @uses_fabric
    def create_deployer(self):
        """
        Creates a deployment user with a directory for Apache configurations.
        """
        user = self.deploy_user
        sudo('useradd -d {0} -m -s /bin/bash {1}'.format(self.deploy_user_home, user))
        sudo('mkdir {0}/.ssh'.format(self.deploy_user_home), user=user)

    @uses_fabric
    def update_deployer_keys(self):
        """
        Replaces deployer keys with the current sysadmin users keys.
        """
        user = self.deploy_user
        file_ = '{0}/.ssh/authorized_keys2'.format(self.deploy_user_home)
        if files.exists(file_):
            sudo('rm {0}'.format(file_), user=user)
        sudo('touch {0}'.format(file_), user=user)
        for _, key_file in self._get_users():
            files.append(file_, open(key_file).read().strip(), use_sudo=True)

    def setup(self):
        """
        Creates sysadmin users and secures the required directories.
        """
        super(BaseInstance, self).setup()
        self.instance.modify_attribute('blockDeviceMapping', ['%s=true' % v[0] for v in self.volume_info] )
        self.create_users(self._get_users())
        self.setup_sudoers()
        # needed for SSH agent forwarding during replication setup:
        self.reset_authentication()
        self.create_deployer()
        self.update_deployer_keys()
        self.bind_app_directories(self.app_dirs, self.app_root)
        self.setup_swap() # after app (potentially secure) partition is created
        self.upgrade_packages()
        self.install_packages(['ntp']) # keep date current


class SessionMixin(RedisMixin):
    """Mixin that creates a session store using Redis."""

    redis_bind = '' # allow connections on all interfaces

    def __init__(self, *args, **kwargs):
        super(SessionMixin, self).__init__(*args, **kwargs)
        self.app_dirs.append('/var/lib/redis')


class CacheMixin(MemcachedMixin):
    """Mixin that creates a cache using Memcached."""

    memcached_bind = '' # allow connections on all interfaces
    memcached_connections = 10000
    memcached_ulimit = 20000
    memcached_threads = 5

    def __init__(self, *args, **kwargs):
        super(CacheMixin, self).__init__(*args, **kwargs)

    @property
    @uses_fabric
    def memcached_memory(self):
        """Returns half of total server memory, in MB"""
        return self.server_memory/2


class QueueMixin(RabbitMqMixin):
    """Mixin that creates a RabbitMQ user and host based on the Fabric env."""

    def __init__(self, *args, **kwargs):
        super(QueueMixin, self).__init__(*args, **kwargs)
        self.app_dirs.append('/var/lib/rabbitmq')

    def setup(self):
        """Create the RabbitMQ user and vhost."""

        super(QueueMixin, self).setup()
        self.create_mq_user(env.deploy_user, env.broker_password)
        self.create_mq_vhost(env.vhost)
        self.set_mq_vhost_permissions(env.vhost, env.deploy_user, '".*" ".*" ".*"')


class DbMixin(PostgresMixin):
    """Mixin that creates a database based on the Fabric env."""

    def __init__(self, *args, **kwargs):
        kwargs['db_settings'] = env.db_settings
        super(DbMixin, self).__init__(*args, **kwargs)
        self.app_dirs.append('/var/lib/postgresql')

    def setup(self):
        """Create the Postgres user and database based on the Fabric env."""

        super(DbMixin, self).setup()
        self.pg_allow_replication('%s_repl' % env.database_user,
                                  env.database_password,
                                  self.postgresql_networks)


class DbMasterMixin(DbMixin):
    """Mixin that creates a database based on the Fabric env."""

    def setup(self):
        """Create the Postgres user and database based on the Fabric env."""

        super(DbMasterMixin, self).setup()
        self.create_db_user(env.database_user, password=env.database_password)
        self.create_db(env.database_name, owner=env.database_user)


class DbSlaveMixin(DbMixin):
    """Mixin that creates a database based on the Fabric env."""

    def setup(self):
        """Create the Postgres user and database based on the Fabric env."""

        super(DbSlaveMixin, self).setup()
        if len(env.servers['db-master']) > 0:
            master = env.servers['db-master'][0]
            self.pg_copy_master(master, '%s_repl' % env.database_user,
                                env.database_password)


class AppMixin(PythonMixin):
    """
    Mixin that installs the Python application dependencies, including the appropriate
    versions of Python, PIP, and virtualenv.
    """

    python_packages = ['python2.7', 'python2.7-dev']
    python_pip_version = '9.0.3'
    python_virtualenv_version = '15.2.0'

    @uses_fabric
    def install_less_and_yuglify(self):
        """
        Adds apt repo for node.js and NPM, installs NPM, and uses NPM to globally
        install the ``lessc`` binary used by django_compressor and pipeline
        and the ``yuglify`` binary used by pipeline.
        """

        node_version = getattr(env, 'node_version', '6.x')
        sudo('curl -sL https://deb.nodesource.com/setup_%s | bash -' % node_version)
        sudo('apt-get -qq -y install nodejs')
        less_version = getattr(env, 'less_version', '1.3.3')
        sudo('npm install -g less@%s' % less_version)
        sudo('npm install -g yuglify')

    @uses_fabric
    def install_system_packages(self):
        """Installs the system packages specified in the environment."""

        # Install required system packages for deployment, plus some extras
        packages = set(env.app_server_packages) | set([
            'supervisor',
            'pgbouncer',
            'stunnel4',
        ])
        self.install_packages(packages)
        # supervisord doesn't start automatically on all Ubuntu versions,
        # so make sure it's started here
        sudo('service supervisor restart')

    @uses_fabric
    def create_webserver_user(self):
        """Create a user for gunicorn, celery, etc."""

        if env.webserver_user != env.deploy_user: # deploy_user already exists
            sudo('useradd --system %(webserver_user)s' % env)

    def setup(self):
        """
        Creates necessary directories, installs required packages, and copies
        the required SSH keys to the server.
        """

        super(AppMixin, self).setup()
        self.install_less_and_yuglify()
        self.install_system_packages()
        self.create_webserver_user()


class WorkerMixin(AppMixin):
    """Mixin that creates a web application server."""

    swap_multiplier = 4 # create 4x the server's RAM in swap

    def __init__(self, *args, **kwargs):
        super(WorkerMixin, self).__init__(*args, **kwargs)


class WebMixin(AppMixin):
    """Mixin that creates a web application server."""

    swap_multiplier = 0 # no swap on the web servers (unhealthy servers will be replaced by Auto Scaling)

    def __init__(self, *args, **kwargs):
        super(WebMixin, self).__init__(*args, **kwargs)

    def setup(self):
        """Installs nginx."""

        super(WebMixin, self).setup()
        self.install_packages(['nginx'])


class CacheInstance(QueueMixin, CacheMixin, SessionMixin, BaseInstance):
    pass


class DbMasterInstance(DbMasterMixin, BaseInstance):
    pass


class DbSlaveInstance(DbSlaveMixin, BaseInstance):
    pass


class WorkerInstance(WorkerMixin, BaseInstance):
    pass


class WebInstance(WebMixin, BaseInstance):
    pass


class CombinedInstance(DbMasterMixin, WebMixin, BaseInstance):
    pass




from fabric.api import *
from fabric.contrib import files

from fabulaws.decorators import *
from fabulaws.api import *
from fabulaws.ubuntu.packages.base import AptMixin

class RedisMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that installs and configures Redis.
    """
    package_name = 'redis'
    redis_packages = ['redis-server']
    redis_bind = '127.0.0.1' # set to '' to bind to all interfaces
    redis_loglevel = 'notice'
    redis_keepalive = 60
    redis_conf = '/etc/redis/redis.conf'
    redis_bind_pattern = '#?\s*bind 127\.0\.0\.1'
    redis_loglevel_pattern = '#?\s*loglevel \w+'
    redis_keepalive_pattern = '#?\s*tcp-keepalive \w+'

    @uses_fabric
    def redis_service(self, cmd):
        sudo('service redis-server {0}'.format(cmd), pty=False) # must pass pty=False

    @uses_fabric
    def redis_configure(self, bind=None, loglevel=None, keepalive=None):
        if bind is None:
            bind = self.redis_bind
        if loglevel is None:
            loglevel = self.redis_loglevel
        if keepalive is None:
            keepalive = self.redis_keepalive
        if bind == '': # all interfaces
            files.comment(self.redis_conf, self.redis_bind_pattern,
                          use_sudo=True)
        else:
            files.sed(self.redis_conf, self.redis_bind_pattern, bind,
                      use_sudo=True)
        loglevel = 'loglevel {0}'.format(loglevel)
        files.sed(self.redis_conf, self.redis_loglevel_pattern, loglevel,
                  use_sudo=True)
        keepalive = 'tcp-keepalive {}'.format(keepalive)
        files.sed(self.redis_conf, self.redis_keepalive_pattern, keepalive,
                  use_sudo=True)
        self.redis_service('restart')

    def bind_app_directories(self, *args, **kwargs):
        # make sure we stop first in case we're being moved to a secure directory
        self.redis_service('stop')
        super(RedisMixin, self).bind_app_directories(*args, **kwargs)
        self.redis_service('start')

    def setup(self):
        """Redis mixin"""

        super(RedisMixin, self).setup()
        self.redis_configure()



from fabulaws.ubuntu.packages.base import AptMixin


class Fail2banMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that installs the fail2ban server for rejecting
    connections from IPs with a history of failed login attemps.
    """
    package_name = 'fail2ban'
    fail2ban_packages = ['fail2ban']



from fabulaws.ubuntu.instances.base import UbuntuInstance

__all__ = ['MicroLucidInstance', 'SmallLucidInstance']


class MicroLucidInstance(UbuntuInstance):
    """
    Sample Micro instance running Ubuntu Lucid on an EBS root filesystem.
    """
    ami = 'ami-3e02f257' # Lucid EBS
    instance_type = 't1.micro'
    key_prefix = 'micro-ubuntu-'


class SmallLucidInstance(UbuntuInstance):
    """
    Sample Small instance running Ubuntu Lucid on on an instance-store root
    filesystem.
    """
    ami = 'ami-7000f019' # Lucid instance store
    instance_type = 'm1.small'
    key_prefix = 'small-ubuntu-'



from fabric.api import *

from fabulaws.decorators import *
from fabulaws.ubuntu.packages.base import AptMixin


class MongoDbMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that installs and configures MongoDB from the
    distro-supplied apt package ``mongodb-server``.
    """
    package_name = 'mongodb'
    mongodb_packages = ['mongodb-server']

    @uses_fabric
    def mongodb_service(self, cmd):
        sudo('service mongodb {0}'.format(cmd))

    def bind_app_directories(self, *args, **kwargs):
        # make sure we stop first in case we're being moved to a secure directory
        self.mongodb_service('stop')
        super(MongoDbMixin, self).bind_app_directories(*args, **kwargs)
        self.mongodb_service('start')


class MongoDb10genMixin(MongoDbMixin):
    """
    FabulAWS Ubuntu mixin that installs and configures MongoDB from the 10gen
    apt repository, which tends to be more up to date.
    """
    mongodb_aptrepo = ('http://downloads-distro.mongodb.org/repo/ubuntu-upstart', 'dist', '10gen', '7F0CEB10')
    mongodb_packages = ['mongodb-10gen']



from fabric.api import *
from fabric.contrib import files

from fabulaws.decorators import uses_fabric
from fabulaws.ubuntu.packages.base import AptMixin

class ShorewallMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that installs and configures the Shorewall firewall
    """
    package_name = 'shorewall'
    shorewall_packages = ['shorewall']
    shorewall_open_ports = ['SSH']
    shorewall_custom = []
    shorewall_allow_icmp = False

    def _get_rules(self, ports):
        for p in ports:
            try:
                p = int(p)
                yield 'ACCEPT net $FW tcp {0}'.format(p)
            except ValueError:
                # looks like a macro, use this format instead
                yield '{0}/ACCEPT net $FW'.format(p)

    @uses_fabric
    def _setup_firewall(self):
        """
        Configures and starts up a Shorewall firewall on the remote server.
        """
        rules = list(self._get_rules(self.shorewall_open_ports))
        rules.extend(self.shorewall_custom)
        with cd('/etc/shorewall'):
            sudo('rsync -a /usr/share/doc/shorewall/examples/one-interface/ .')
            if files.exists('shorewall.conf.gz'):
                sudo('gunzip -f shorewall.conf.gz')
            files.append('rules', '\n'.join(rules), use_sudo=True)
            sudo('sed -i "s/STARTUP_ENABLED=No/STARTUP_ENABLED=Yes/" shorewall.conf')
        if self.shorewall_allow_icmp:
            files.sed('/etc/shorewall/rules', r'Ping\(DROP\)',
                      'Ping(ACCEPT)', use_sudo=True)
        sudo('shorewall start')

    def setup(self):
        """
        Hooks into the FabulAWS setup process to configure Shorewall on the
        server.
        """
        super(ShorewallMixin, self).setup()
        self._setup_firewall()



import os
import time
import copy
import logging

import boto.exception

from decimal import Decimal

from fabric.api import *
from fabric.contrib import files

from fabulaws.api import *
from fabulaws.decorators import uses_fabric
from fabulaws.ec2 import EC2Instance
from fabulaws.ubuntu.packages.base import BaseAptMixin

__all__ = ['UbuntuInstance']

logger = logging.getLogger('fabulaws.ubuntu.instances.base')

class UbuntuInstance(BaseAptMixin, EC2Instance):
    """
    Base class for all Ubuntu instances.
    """
    user = 'ubuntu'
    admin_groups = ['admin']
    volume_info = [] # tuples of (device, mount_point, size_in_GB, type, passwd)
    fs_type = 'ext3'
    fs_encrypt = False
    ubuntu_mirror = None
    mount_script = '/mount-deferred.sh'

    def __init__(self, *args, **kwargs):
        self.volumes = []
        super(UbuntuInstance, self).__init__(*args, **kwargs)

    @uses_fabric
    def _wait_for_device(self, device, max_tries=30):
        """
        Waits for the given device to manifest in /dev, and returns the actual
        path of the device.
        """
        # try all possible paths for the device in turn until one shows up
        devices = [device, device.replace('/dev/sd', '/dev/xvd')]
        logger.info('Waiting for device {0} to appear'.format(device))
        for _ in range(max_tries):
            for device in devices:
                if files.exists(device):
                    logger.debug('Found device {0}'.format(device))
                    return device
            time.sleep(1)

    @uses_fabric
    def _encrypt_device(self, device, passwd=None):
        """
        Encrypts the given device.  If a password is not specified,
        a prompt will be issued.
        """
        if not passwd:
            passwd = getpass.getpass('Enter LUKS passphrase for cryptsetup: ')
        self.install_packages(['python-pexpect', 'cryptsetup'])
        crypt = 'crypt-{0}'.format(device.split('/')[-1])
        with hide('stdout'):
            answers = [
                (r'Are you sure\? \(Type uppercase yes\):', 'YES'),
                ('Enter .*passphrase:', passwd),
                ('Verify passphrase:', passwd),
            ]
            answer_sudo('cryptsetup -y luksFormat {device}'.format(device=device),
                        answers=answers)
            answers = [('Enter passphrase for .+:', passwd)]
            answer_sudo('cryptsetup luksOpen {device} {crypt}'
                        ''.format(device=device, crypt=crypt),
                        answers=answers)
        device = '/dev/mapper/{0}'.format(crypt)
        return device

    @uses_fabric
    def _mount_and_persist(self, device, mount_point, fs_type=None, opts=None):
        mount_cmd = ' '.join([
            'mount',
            '-o {}'.format(opts) if opts else '',
            '-t {}'.format(fs_type) if fs_type else '',
            device,
            mount_point,
        ])
        fstab_entry = ' '.join([
            device,
            mount_point,
            fs_type if fs_type else self.fs_type,
            opts if opts else 'defaults',
            '0',
            '0',
        ])
        sudo(mount_cmd)
        if self.fs_encrypt:
            if not files.exists(self.mount_script):
                files.append(self.mount_script, "#!/bin/sh", use_sudo=True)
                sudo('chmod a+x {}'.format(self.mount_script))
            files.append(self.mount_script, mount_cmd, use_sudo=True)
        else:
            files.append('/etc/fstab', fstab_entry, use_sudo=True)

    @uses_fabric
    def _format_volume(self, device, mount_point, passwd=None):
        """
        Creates an EBS volume of size ``vol_size``, manifests the device at
        ``device`` in this instance, and mounts it at ``mount_point``.
        """
        logger.info('Formatting {0}'.format(device))
        if self.fs_encrypt:
            device = self._encrypt_device(device, passwd)
        sudo('mkfs.{0} {1}'.format(self.fs_type, device))
        sudo('mkdir {0}'.format(mount_point))
        self._mount_and_persist(device, mount_point)

    def _set_volume_tags(self, vol, device, tags=None):
        if tags is None:
            tags = copy.copy(self._tags) or {}
        if 'Name' in tags:
            tags['Name'] = '_'.join([tags['Name'], os.path.basename(device)])
        else:
            tags['Name'] = '_'.join([inst.id, os.path.basename(device)])
        tags['device'] = device
        self.conn.create_tags([vol.id], tags)

    def add_tags(self, tags):
        """
        Propagate any tag updates on the instance to the associated volumes
        (e.g., for name changes).
        """
        super(UbuntuInstance, self).add_tags(tags)
        # allow super class to update self._tags and use that in _set_volume_tags
        for vol, (device, _, _, _, _)  in zip(self.volumes, self.volume_info):
            self._set_volume_tags(vol, device)

    @uses_fabric
    def _create_volume(self, device, mount_point, vol_size, vol_type, passwd=None):
        """
        Creates an EBS volume of size ``vol_size``, manifests the device at
        ``device`` in this instance, and mounts it at ``mount_point``.
        """
        logger.info('Attaching {0}'.format(device))
        inst = self.instance
        # the placement is the availability zone
        vol = self.conn.create_volume(vol_size, inst.placement, volume_type=vol_type)
        self._set_volume_tags(vol, device)
        try:
            logger.debug('Waiting for volume {0} to become AVAILABLE '
                         '(current state={1})'.format(vol.id, vol.volume_state()))
            while vol.volume_state() != 'available':
                time.sleep(1)
                vol.update()
                logger.debug('  current state: {0}'.format(vol.volume_state()))
            vol.attach(inst.id, device)
            logger.debug('Waiting for volume {0} to become ATTACHED '
                         '(current state={1})'.format(vol.id, vol.volume_state()))
            while vol.volume_state() != 'in-use':
                time.sleep(1)
                vol.update()
                logger.debug('  current state: {0}'.format(vol.volume_state()))
        except:
            self._destroy_volume(vol)
            raise
        return vol

    def _destroy_volume(self, vol):
        """
        Forcibly detaches and destroys the given EBS volume, where vol is an
        instance of ``boto.ec2.volume``.
        """
        logger.debug('Detaching volume {0}'.format(vol.id))
        try:
            vol.detach(force=True)
        except boto.exception.EC2ResponseError:
            logger.exception('Failed to detach volume; continuing anyway.')
        logger.debug('Waiting for volume {0} to become '
                     'available'.format(vol.id))
        while vol.volume_state() == 'in-use':
            time.sleep(1)
            vol.update()
        logger.debug('Deleting volume {0}'.format(vol.id))
        vol.delete()

    @property
    @uses_fabric
    def server_memory(self):
        """Returns total server memory, in MB"""
        if not hasattr(self, '_server_memory'):
            mem = run('cat /proc/meminfo|grep MemTotal')
            self._server_memory = int(mem.split()[1]) / 1024
        return self._server_memory

    @property
    @uses_fabric
    def ubuntu_release(self):
        if not hasattr(self, '_ubuntu_release'):
            self._ubuntu_release = Decimal(run('lsb_release -r -s').strip())
        return self._ubuntu_release

    @uses_fabric
    def setup_mirror(self, mirror=None):
        if not mirror:
            mirror = self.ubuntu_mirror
        if mirror:
            orig = '{region}.ec2.archive.ubuntu.com'.format(region=self.region)
            mirror = mirror.format(region=self.region)
            files.sed('/etc/apt/sources.list', orig, mirror, use_sudo=True)

    def setup(self):
        """
        Extends the base EC2Instance ``setup()`` method with routines to
        update apt sources on the instance.  Also creates volumes defined in
        the ``volume_info`` list on this instance.
        """
        super(UbuntuInstance, self).setup()
        # this is required because we may need to install cryptsetup when
        # creating volumes
        self.setup_mirror()
        self.update_apt_sources()
        # the first apt-get update may update sources.list, so re-run it here
        self.setup_mirror()
        self.update_apt_sources()
        for vol in self.volume_info:
            if len(vol) == 5:
                device, mount_point, vol_size, vol_type, passwd = vol
            else:
                raise Exception('volume_info must be populated with tuples of '
                                '(device, mount_point, vol_size, vol_type, passwd)')
            if device == 'instance-store':
                with self:
                    device = run("mount|grep /mnt|cut -d' ' -f1").strip()
                    sudo('umount {0}'.format(device))
            else:
                self.volumes.append(self._create_volume(device, mount_point,
                                                        vol_size, vol_type, passwd))
            device = self._wait_for_device(device)
            self._format_volume(device, mount_point, passwd)

    @uses_fabric
    def create_users(self, users, ignore_existing=True):
        """
        Create admin users and deploy SSH keys to the server.  ``users`` is
        a list of (username, keyfile) tuples.  The users will be created with
        empty passwords.
        """
        if self.admin_groups:
            groups = '-G {0}'.format(','.join(self.admin_groups))
        else:
            groups = ''
        for name, keyfile in users:
            if ignore_existing and files.exists('/home/{0}'.format(name)):
                logger.info('Not creating existing user {0}'.format(name))
            else:
                sudo('useradd -m {0} -s /bin/bash {1}'.format(groups, name))
                sudo('passwd -d {0}'.format(name))
                sudo('mkdir /home/{0}/.ssh'.format(name))
            put(keyfile, '/home/{0}/.ssh/authorized_keys2'.format(name),
                use_sudo=True, mode=0600)
            sudo('chown -R {0} /home/{0}/.ssh'.format(name))
            # if a file exists with the comment field (e.g., for GECOS info)
            # for the user, use usermod -c to add it.
            gecos_file = keyfile + '.gecos'
            if os.path.exists(gecos_file):
                with open(gecos_file, 'r') as gecos_fd:
                    gecos = gecos_fd.readline().strip()
                    sudo('usermod -c "{}" {}'.format(gecos, name))

    @uses_fabric
    def bind_app_directories(self, app_dirs, app_root):
        """
        Move the given directories, ``app_dirs'', to the specified file system
        mounted at ``app_root'' (optionally secure if ``self.fs_encrypt == True'').
        """
        assert files.exists(app_root)
        for app_dir in app_dirs:
            bound_app_dir = ''.join([app_root, app_dir])
            bound_parent_dir = call_python('os.path.dirname', bound_app_dir)
            if files.exists(app_dir):
                sudo('mkdir -p {0}'.format(bound_parent_dir))
                sudo('mv {0} {1}'.format(app_dir, bound_app_dir))
            else:
                sudo('mkdir -p {0}'.format(bound_app_dir))
            sudo('mkdir -p {0}'.format(app_dir))
            self._mount_and_persist(bound_app_dir, app_dir, fs_type='none', opts='bind')

    def cleanup(self):
        """
        If needed, destroys any volumes created for this instance and then
        calls the base class's ``cleanup()`` method.
        """
        if self._terminate:
            while self.volumes:
                vol = self.volumes.pop()
                self._destroy_volume(vol)
        else:
            for vol in self.volumes:
                logger.warning('Left volume "{0}" in state "{1}"'
                               ''.format(vol.id, vol.volume_state()))
        super(UbuntuInstance, self).cleanup()
